import sys

from ...exceptions import FastAgencyCLIPythonVersionError
from ...helpers import check_imports

if sys.version_info < (3, 10):
    raise FastAgencyCLIPythonVersionError("Mesop requires Python 3.10 or higher")

check_imports(["mesop"], "mesop")

from .mesop import MesopUI  # noqa: E402

__all__ = ["MesopUI"]



from functools import wraps
from typing import Any

from fastapi_code_generator.parser import OpenAPIParser

from ...logging import get_logger

logger = get_logger(__name__)


def patch_parse_schema() -> None:
    org_parse_schema = OpenAPIParser.parse_schema

    @wraps(org_parse_schema)
    def my_parse_schema(*args: Any, **kwargs: Any) -> Any:
        data_type = org_parse_schema(*args, **kwargs)
        if data_type.reference and data_type.reference.duplicate_name:
            data_type.reference.name = data_type.reference.duplicate_name
        return data_type

    OpenAPIParser.parse_schema = my_parse_schema
    logger.info("Patched OpenAPIParser.parse_schema")



import re
from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass, field, fields
from typing import Any, Literal, Optional, Protocol, Type
from uuid import uuid4

from .logging import get_logger

__all__ = [
    "IOMessage",
    "TextMessage",
    "SuggestedFunctionCall",
    "FunctionCallExecution",
    "TextInput",
    "MultipleChoice",
    "SystemMessage",
    "KeepAlive",
    "WorkflowStarted",
    "WorkflowCompleted",
    "Error",
    "MessageType",
    "MessageProcessorProtocol",
    "MessageProcessorMixin",
]


logger = get_logger(__name__)


def _camel_to_snake(name: str) -> str:
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


# we keep this hardcoded for mypy and type checks
MessageType = Literal[
    "text_message",
    "suggested_function_call",
    "function_call_execution",
    "text_input",
    "multiple_choice",
    "system_message",
    "keep_alive",
    "workflow_started",
    "workflow_completed",
    "error",
]


@dataclass
class IOMessage(
    ABC
):  # `IOMessage` is an abstract base class, but it has no abstract methods
    workflow_uuid: str
    sender: Optional[str] = None
    recipient: Optional[str] = None
    auto_reply: bool = False
    uuid: str = field(default_factory=lambda: str(uuid4().hex))

    @property
    def type(self) -> MessageType:
        retval: MessageType = _camel_to_snake(self.__class__.__name__)  # type: ignore[assignment]
        return retval

    @staticmethod
    def _get_message_class(type: Optional[MessageType]) -> "Type[IOMessage]":
        type = type or "text_message"
        lookup: dict[MessageType, "Type[IOMessage]"] = {
            "text_message": TextMessage,
            "suggested_function_call": SuggestedFunctionCall,
            "function_call_execution": FunctionCallExecution,
            "text_input": TextInput,
            "multiple_choice": MultipleChoice,
            "keep_alive": KeepAlive,
            "system_message": SystemMessage,
            "workflow_started": WorkflowStarted,
            "workflow_completed": WorkflowCompleted,
            "error": Error,
        }
        return lookup[type]

    @staticmethod
    def create(type: Optional[MessageType] = None, **kwargs: Any) -> "IOMessage":
        cls = IOMessage._get_message_class(type)

        content = kwargs.pop("content", {})
        kwargs.update(content)

        return cls(**kwargs)

    @staticmethod
    def _get_parameters_names() -> list[str]:
        return [field.name for field in fields(IOMessage)]

    def model_dump(self) -> dict[str, Any]:
        params_names = IOMessage._get_parameters_names()
        d = asdict(self)
        content = {k: v for k, v in d.items() if k not in params_names}
        retval = {k: v for k, v in d.items() if k in params_names}
        retval["content"] = content
        retval["type"] = self.type
        return retval


# message type that asks user something


@dataclass
class AskingMessage(IOMessage): ...


# type of output messages
@dataclass
class TextMessage(IOMessage):
    body: Optional[str] = None

    def __post_init__(self) -> None:
        """Set the default value for the `type` attribute."""
        if self.type is None:
            self.type = "text_message"


@dataclass
class SuggestedFunctionCall(IOMessage):
    function_name: Optional[str] = None
    call_id: Optional[str] = None
    arguments: dict[str, Any] = field(default_factory=dict)


@dataclass
class FunctionCallExecution(IOMessage):
    function_name: Optional[str] = None
    call_id: Optional[str] = None
    retval: Any = None


# types of input messages
@dataclass
class TextInput(AskingMessage):
    prompt: Optional[str] = None
    suggestions: list[str] = field(default_factory=list)
    password: bool = False


@dataclass
class MultipleChoice(AskingMessage):
    prompt: Optional[str] = None
    choices: list[str] = field(default_factory=list)
    default: Optional[str] = None
    single: bool = True
    # todo: add validation


@dataclass
class SystemMessage(IOMessage):
    message: dict[str, Any] = field(default_factory=dict)


@dataclass
class WorkflowStarted(IOMessage):
    name: Optional[str] = None
    description: Optional[str] = None
    params: dict[str, Any] = field(default_factory=dict)


@dataclass
class WorkflowCompleted(IOMessage):
    result: Optional[str] = None


@dataclass
class Error(IOMessage):
    short: Optional[str] = None
    long: Optional[str] = None


@dataclass
class KeepAlive(IOMessage): ...


class MessageProcessorProtocol(Protocol):
    def process_message(self, message: IOMessage) -> Optional[str]: ...

    def text_message(
        self,
        # common parameters for all messages
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # text_message specific parameters
        body: Optional[str] = None,
    ) -> Optional[str]: ...

    def suggested_function_call(
        self,
        # common parameters for all messages
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # suggested_function_call specific parameters
        function_name: Optional[str] = None,
        call_id: Optional[str] = None,
        arguments: Optional[dict[str, Any]] = None,
    ) -> Optional[str]: ...

    def function_call_execution(
        self,
        # common parameters for all messages
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # function_call_execution specific parameters
        function_name: Optional[str] = None,
        call_id: Optional[str] = None,
        retval: Any = None,
    ) -> Optional[str]: ...

    def text_input(
        self,
        # common parameters for all messages
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # text_input specific parameters
        prompt: Optional[str] = None,
        suggestions: Optional[list[str]] = None,
        password: bool = False,
    ) -> Optional[str]: ...

    def multiple_choice(
        self,
        # common parameters for all messages
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # multiple_choice specific parameters
        prompt: Optional[str] = None,
        choices: Optional[list[str]] = None,
        default: Optional[str] = None,
        single: bool = True,
    ) -> Optional[str]: ...

    def system_message(
        self,
        # common parameters for all messages
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # system_message specific parameters
        message: Optional[dict[str, Any]] = None,
    ) -> Optional[str]: ...

    def workflow_started(
        self,
        # common parameters for all messages
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # workflow_started specific parameters
        name: Optional[str] = None,
        description: Optional[str] = None,
        params: Optional[dict[str, Any]] = None,
    ) -> Optional[str]: ...

    def workflow_completed(
        self,
        # common parameters for all messages
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # workflow_completed specific parameters
        result: Optional[str] = None,
    ) -> Optional[str]: ...

    def error(
        self,
        # common parameters for all messages
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # error specific parameters
        short: Optional[str] = None,
        long: Optional[str] = None,
    ) -> Optional[str]: ...

    def keep_alive(
        self,
        # common parameters for all messages
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
    ) -> Optional[str]: ...


class MessageProcessorMixin(ABC):
    def visit(self, message: IOMessage) -> Optional[str]:
        method_name = f"visit_{message.type}"
        method = getattr(self, method_name, self.visit_default)
        return method(message)

    @abstractmethod
    def visit_default(self, message: IOMessage) -> Optional[str]: ...

    def visit_text_message(self, message: TextMessage) -> Optional[str]:
        return self.visit_default(message)

    def visit_suggested_function_call(
        self, message: SuggestedFunctionCall
    ) -> Optional[str]:
        return self.visit_default(message)

    def visit_function_call_execution(
        self, message: FunctionCallExecution
    ) -> Optional[str]:
        return self.visit_default(message)

    def visit_text_input(self, message: TextInput) -> Optional[str]:
        return self.visit_default(message)

    def visit_multiple_choice(self, message: MultipleChoice) -> Optional[str]:
        return self.visit_default(message)

    def visit_system_message(self, message: SystemMessage) -> Optional[str]:
        return self.visit_default(message)

    def visit_keep_alive(self, message: KeepAlive) -> Optional[str]:
        return self.visit_default(message)

    def visit_workflow_started(self, message: WorkflowStarted) -> Optional[str]:
        return self.visit_default(message)

    def visit_workflow_completed(self, message: WorkflowCompleted) -> Optional[str]:
        return self.visit_default(message)

    def visit_error(self, message: Error) -> Optional[str]:
        return self.visit_default(message)

    def process_message(self, message: IOMessage) -> Optional[str]:
        try:
            return self.visit(message)
        except Exception as e:
            # log the error and return None
            logger.error(f"Error processing message ({message}): {e}", stack_info=True)
            return None

    def text_message(
        self,
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        body: Optional[str] = None,
    ) -> Optional[str]:
        uuid = uuid or str(uuid4().hex)
        return self.process_message(
            TextMessage(
                sender=sender,
                recipient=recipient,
                auto_reply=auto_reply,
                uuid=uuid,
                workflow_uuid=workflow_uuid,
                body=body,
            )
        )

    def suggested_function_call(
        self,
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        function_name: Optional[str] = None,
        call_id: Optional[str] = None,
        arguments: Optional[dict[str, Any]] = None,
    ) -> Optional[str]:
        uuid = uuid or str(uuid4().hex)
        arguments = arguments or {}
        return self.process_message(
            SuggestedFunctionCall(
                sender=sender,
                recipient=recipient,
                auto_reply=auto_reply,
                uuid=uuid,
                workflow_uuid=workflow_uuid,
                function_name=function_name,
                call_id=call_id,
                arguments=arguments,
            )
        )

    def function_call_execution(
        self,
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        function_name: Optional[str] = None,
        call_id: Optional[str] = None,
        retval: Any = None,
    ) -> Optional[str]:
        uuid = uuid or str(uuid4().hex)
        return self.process_message(
            FunctionCallExecution(
                sender=sender,
                recipient=recipient,
                auto_reply=auto_reply,
                uuid=uuid,
                workflow_uuid=workflow_uuid,
                function_name=function_name,
                call_id=call_id,
                retval=retval,
            )
        )

    def text_input(
        self,
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        prompt: Optional[str] = None,
        suggestions: Optional[list[str]] = None,
        password: bool = False,
    ) -> Optional[str]:
        uuid = uuid or str(uuid4().hex)
        suggestions = suggestions or []
        return self.process_message(
            TextInput(
                sender=sender,
                recipient=recipient,
                auto_reply=auto_reply,
                uuid=uuid,
                workflow_uuid=workflow_uuid,
                prompt=prompt,
                suggestions=suggestions,
                password=password,
            )
        )

    def multiple_choice(
        self,
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        prompt: Optional[str] = None,
        choices: Optional[list[str]] = None,
        default: Optional[str] = None,
        single: bool = True,
    ) -> Optional[str]:
        uuid = uuid or str(uuid4().hex)
        choices = choices or []
        return self.process_message(
            MultipleChoice(
                sender=sender,
                recipient=recipient,
                auto_reply=auto_reply,
                uuid=uuid,
                workflow_uuid=workflow_uuid,
                prompt=prompt,
                choices=choices,
                default=default,
                single=single,
            )
        )

    def system_message(
        self,
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        message: Optional[dict[str, Any]] = None,
    ) -> Optional[str]:
        uuid = uuid or str(uuid4().hex)
        message = message or {}
        return self.process_message(
            SystemMessage(
                sender=sender,
                recipient=recipient,
                auto_reply=auto_reply,
                uuid=uuid,
                workflow_uuid=workflow_uuid,
                message=message,
            )
        )

    def workflow_started(
        self,
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        params: Optional[dict[str, Any]] = None,
    ) -> Optional[str]:
        uuid = uuid or str(uuid4().hex)
        params = params or {}
        return self.process_message(
            WorkflowStarted(
                sender=sender,
                recipient=recipient,
                auto_reply=auto_reply,
                uuid=uuid,
                workflow_uuid=workflow_uuid,
                name=name,
                description=description,
                params=params,
            )
        )

    def workflow_completed(
        self,
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        result: Optional[str] = None,
    ) -> Optional[str]:
        uuid = uuid or str(uuid4().hex)
        return self.process_message(
            WorkflowCompleted(
                sender=sender,
                recipient=recipient,
                auto_reply=auto_reply,
                uuid=uuid,
                workflow_uuid=workflow_uuid,
                result=result,
            )
        )

    def error(
        self,
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        short: Optional[str] = None,
        long: Optional[str] = None,
    ) -> Optional[str]:
        uuid = uuid or str(uuid4().hex)
        return self.process_message(
            Error(
                sender=sender,
                recipient=recipient,
                auto_reply=auto_reply,
                uuid=uuid,
                workflow_uuid=workflow_uuid,
                short=short,
                long=long,
            )
        )

    def keep_alive(
        self,
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
    ) -> Optional[str]:
        uuid = uuid or str(uuid4().hex)
        return self.process_message(
            KeepAlive(
                sender=sender,
                recipient=recipient,
                auto_reply=auto_reply,
                uuid=uuid,
                workflow_uuid=workflow_uuid,
            )
        )



import threading
import time
import traceback
from collections.abc import Generator, Iterator
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from queue import Queue
from tempfile import TemporaryDirectory
from typing import Any, Callable, ClassVar, Optional
from uuid import uuid4

import mesop as me
from mesop.bin.bin import FLAGS as MESOP_FLAGS
from mesop.bin.bin import main as mesop_main

from ...base import (
    UI,
    CreateWorkflowUIMixin,
    ProviderProtocol,
    Runnable,
)
from ...logging import get_logger
from ...messages import (
    AskingMessage,
    IOMessage,
    KeepAlive,
    MessageProcessorMixin,
    MultipleChoice,
    TextInput,
    TextMessage,
    WorkflowCompleted,
)
from .styles import MesopHomePageStyles
from .timer import configure_static_file_serving

logger = get_logger(__name__)


@dataclass
class MesopMessage:
    """A Mesop message."""

    io_message: IOMessage
    conversation: "MesopUI"


class MesopUI(MessageProcessorMixin, CreateWorkflowUIMixin):  # UIBase
    _import_string: Optional[str] = None
    _main_path: Optional[str] = None
    _created_instance: Optional["MesopUI"] = None
    _app: Optional[Runnable] = None
    _me: Optional[Callable[..., Any]] = None

    def __init__(
        self,
        super_conversation: "Optional[MesopUI]" = None,
        *,
        security_policy: Optional[me.SecurityPolicy] = None,
        styles: Optional[MesopHomePageStyles] = None,
        keep_alive: Optional[bool] = False,
    ) -> None:
        """Initialize the console UI object.

        Args:
            super_conversation (Optional[MesopUI], optional): The super conversation. Defaults to None.
            security_policy (Optional[me.SecurityPolicy], optional): The security policy. Defaults to None.
            styles (Optional[MesopHomePageStyles], optional): The styles. Defaults to None.
            keep_alive (Optional[bool]): If keep alive messages should be inserted, defaults to False`
        """
        logger.info(f"Initializing MesopUI: {self}")
        try:
            self.id: str = uuid4().hex
            self.super_conversation: Optional[MesopUI] = super_conversation
            self.sub_conversations: list[MesopUI] = []
            self._in_queue: Optional[Queue[str]] = None
            self._out_queue: Optional[Queue[MesopMessage]] = None

            self._keep_me_alive = keep_alive
            self._keep_alive_thread: Optional[threading.Thread] = None
            if super_conversation is None:
                self._in_queue = Queue()
                self._out_queue = Queue()
                self.keep_me_alive()

            MesopUI.register(self)

            if MesopUI._me is None:
                from .main import create_home_page, me

                create_home_page(self, security_policy=security_policy, styles=styles)
                MesopUI._me = me

        except Exception as e:
            logger.error(e, exc_info=True)
            raise
        logger.info(f"Initialized MesopUI: {self}")

    _registry: ClassVar[dict[str, "MesopUI"]] = {}

    def keep_me_alive(self) -> None:
        def keep_alive_worker() -> None:
            while self._keep_me_alive:
                time.sleep(3)
                if self._out_queue:
                    # todo: do something more elegant
                    msg = KeepAlive(workflow_uuid="")
                    mesop_msg = self._mesop_message(msg)
                    logger.debug(f"putting keepalive {msg.uuid}")
                    self._out_queue.put(mesop_msg)

        if self._keep_me_alive and self._keep_alive_thread is None:
            self._keep_alive_thread = threading.Thread(target=keep_alive_worker)
            self._keep_alive_thread.start()

    def do_not_keep_me_alive(self) -> None:
        self._keep_me_alive = False

    @classmethod
    def get_created_instance(cls) -> "MesopUI":
        created_instance = cls._created_instance
        if created_instance is None:
            raise RuntimeError("MesopUI has not been created yet.")

        return created_instance

    @property
    def app(self) -> Runnable:
        app = MesopUI._app
        if app is None:
            logger.error("MesopUI has not been created yet.")
            raise RuntimeError("MesopUI has not been created yet.")

        return app

    @contextmanager
    def create(self, app: Runnable, import_string: str) -> Iterator[None]:
        logger.info(f"Creating MesopUI with import string: {import_string}")
        MesopUI._app = app
        MesopUI._import_string = import_string

        start_script = """import fastagency.ui.mesop.main"""

        with TemporaryDirectory() as temp_dir:
            main_path = Path(temp_dir) / "main.py"
            with main_path.open("w") as f:
                f.write(start_script)

            MESOP_FLAGS.mark_as_parsed()
            MesopUI._main_path = str(main_path)
            MesopUI._created_instance = self

            yield

    def start(
        self,
        *,
        app: Runnable,
        import_string: str,
        name: Optional[str] = None,
        params: dict[str, Any],
        single_run: bool = False,
    ) -> None:
        logger.info(
            f"Starting MesopUI: import_string={self._import_string}, main_path={self._main_path}"
        )
        if single_run:
            logger.warning("single_run parameter is currently not supported in MesopUI")

        MesopUI._app = app

        mesop_main(["mesop", self._main_path])

    @classmethod
    def register(cls, conversation: "MesopUI") -> None:
        cls._registry[conversation.id] = conversation

    @classmethod
    def get_conversation(cls, id: str) -> "MesopUI":
        return cls._registry[id]

    @classmethod
    def unregister(cls, conversation: "MesopUI") -> None:
        del cls._registry[conversation.id]

    @property
    def is_root_conversation(self) -> bool:
        return self.super_conversation is not None

    @property
    def root_conversation(self) -> "MesopUI":
        if self.super_conversation is None:
            return self
        else:
            return self.super_conversation.root_conversation

    @property
    def in_queue(self) -> Queue[str]:
        queue = self.root_conversation._in_queue
        return queue  # type: ignore[return-value]

    @property
    def out_queue(self) -> Queue[MesopMessage]:
        queue = self.root_conversation._out_queue
        return queue  # type: ignore[return-value]

    @property
    def level(self) -> int:
        return (
            0 if self.super_conversation is None else self.super_conversation.level + 1
        )

    def _publish(self, mesop_msg: MesopMessage) -> None:
        self.out_queue.put(mesop_msg)

    def _mesop_message(self, io_message: IOMessage) -> MesopMessage:
        return MesopMessage(
            conversation=self,
            io_message=io_message,
        )

    def visit_default(self, message: IOMessage) -> None:
        mesop_msg = self._mesop_message(message)
        self._publish(mesop_msg)

    def visit_text_message(self, message: TextMessage) -> None:
        mesop_msg = self._mesop_message(message)
        self._publish(mesop_msg)

    def visit_text_input(self, message: TextInput) -> str:
        mesop_msg = self._mesop_message(message)
        self._publish(mesop_msg)
        return self.in_queue.get()

    def visit_multiple_choice(self, message: MultipleChoice) -> str:
        mesop_msg = self._mesop_message(message)
        self._publish(mesop_msg)
        return self.in_queue.get()

    def process_message(self, message: IOMessage) -> Optional[str]:
        return self.visit(message)

    def create_subconversation(self) -> "MesopUI":
        sub_conversation = MesopUI(self)
        self.sub_conversations.append(sub_conversation)

        return sub_conversation

    def _is_stream_braker(self, message: IOMessage) -> bool:
        return isinstance(message, (AskingMessage, WorkflowCompleted, KeepAlive))

    def respond(self, message: str) -> None:
        self.in_queue.put(message)

    @classmethod
    def respond_to(
        cls, conversation_id: str, message: str
    ) -> Generator[MesopMessage, None, None]:
        conversation = cls.get_conversation(conversation_id)
        conversation.respond(message)
        return conversation.get_message_stream()

    def get_message_stream(self) -> Generator[MesopMessage, None, None]:
        while True:
            message = self.out_queue.get()
            if self._is_stream_braker(message.io_message):
                yield message
                break
            yield message

    def handle_wsgi(
        self,
        app: "Runnable",
        environ: dict[str, Any],
        start_response: Callable[..., Any],
    ) -> list[bytes]:
        logger.debug(f"Starting MesopUI using WSGI interface with app: {app}")
        MesopUI._created_instance = self
        MesopUI._app = app

        if configure_static_file_serving is None:  # pragme: no cover
            logger.error("configure_static_file_serving is None")

        if MesopUI._me is None:
            logger.error("MesopUI._me is None")
            raise RuntimeError("MesopUI._me is None")

        return MesopUI._me(environ, start_response)  # type: ignore[no-any-return]


def run_workflow_mesop(provider: ProviderProtocol, name: str) -> UI:
    def workflow_worker(
        provider: ProviderProtocol, name: str, mesop_ui: MesopUI, workflow_uuid: str
    ) -> None:
        ui = mesop_ui.create_workflow_ui(workflow_uuid)
        try:
            provider.run(
                name=name,
                ui=ui,
            )
        except Exception as e:
            logger.error(
                f"Unexpected exception raised in Mesop workflow worker: {e}",
                stack_info=True,
            )
            ui.error(
                sender="Mesop workflow_worker",
                short=f"Unexpected exception raised: {e}",
                long=traceback.format_exc(),
            )
            return
        finally:
            mesop_ui.do_not_keep_me_alive()

    ui_base = MesopUI(keep_alive=True)
    workflow_uuid = ui_base.id

    # subconversation = ui_base.create_subconversation()
    thread = threading.Thread(
        target=workflow_worker, args=(provider, name, ui_base, workflow_uuid)
    )
    thread.start()

    return ui_base.create_workflow_ui(workflow_uuid)

    # # needed for uvicorn to recognize the class as a valid ASGI application
    # async def __call__(
    #     self,
    #     scope: dict[str, Any],
    #     receive: Callable[[], Awaitable[dict]],
    #     send: Callable[[dict], Awaitable[None]],
    # ) -> None:
    #     MesopUI._created_instance = self
    #     from .main import me

    #     return await me(scope, receive, send)



import json
import re
from collections.abc import Iterable, Mapping
from dataclasses import asdict, dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Optional,
    Protocol,
    Union,
    runtime_checkable,
)

from autogen.agentchat import ConversableAgent
from autogen.io import IOStream

from ...base import (
    UI,
    Workflow,
    WorkflowsProtocol,
    check_register_decorator,
)
from ...logging import get_logger
from ...messages import (
    AskingMessage,
    IOMessage,
    MessageType,
    MultipleChoice,
    TextInput,
)

if TYPE_CHECKING:
    from fastagency.api.openapi import OpenAPI

__all__ = [
    "AutoGenWorkflows",
    "IOStreamAdapter",
]


# Get the logger
logger = get_logger(__name__)

logger.info("Importing autogen.base.py")

_patterns = {
    "end_of_message": (
        "^\\n--------------------------------------------------------------------------------\\n$",
    ),
    "auto_reply": (
        "^\\x1b\\[31m\\n>>>>>>>> USING AUTO REPLY...\\x1b\\[0m\\n$",
        "^\\n>>>>>>>> USING AUTO REPLY...\\n$",
    ),
    "sender_recipient": (
        "^\\x1b\\[33m([a-zA-Z0-9_-]+)\\x1b\\[0m \\(to ([a-zA-Z0-9_-]+)\\):\\n\\n$",
        "^([a-zA-Z0-9_-]+) \\(to ([a-zA-Z0-9_-]+)\\):\\n\\n$",
    ),
    "suggested_function_call": (
        "^\\x1b\\[32m\\*\\*\\*\\*\\* Suggested tool call \\((call_[a-zA-Z0-9]+)\\): ([a-zA-Z0-9_]+) \\*\\*\\*\\*\\*\\x1b\\[0m\\n$",
        "^\\*\\*\\*\\*\\* Suggested tool call \\((call_[a-zA-Z0-9]+)\\): ([a-zA-Z0-9_]+) \\*\\*\\*\\*\\*\\n$",
    ),
    "stars": ("^\\x1b\\[32m(\\*{69}\\*+)\\x1b\\[0m\n$", "^(\\*{69}\\*+)\\n$"),
    "function_call_execution": (
        "^\\x1b\\[35m\\n>>>>>>>> EXECUTING FUNCTION ([a-zA-Z_]+)...\\x1b\\[0m\\n$",
        "^\\n>>>>>>>> EXECUTING FUNCTION ([a-zA-Z_]+)...\\n$",
    ),
    "response_from_calling_tool": (
        "^\\x1b\\[32m\\*\\*\\*\\*\\* Response from calling tool \\((call_[a-zA-Z0-9_]+)\\) \\*\\*\\*\\*\\*\\x1b\\[0m\\n$",
        "^\\*\\*\\*\\*\\* Response from calling tool \\((call_[a-zA-Z0-9_]+)\\) \\*\\*\\*\\*\\*\\n$",
    ),
    "no_human_input_received": (
        "^\\x1b\\[31m\\n>>>>>>>> NO HUMAN INPUT RECEIVED\\.\\x1b\\[0m$",
        "^\\n>>>>>>>> NO HUMAN INPUT RECEIVED\\.$",
    ),
    "user_interrupted": ("^USER INTERRUPTED\\n$",),
    "arguments": ("^Arguments: \\n(.*)\\n$",),
    "auto_reply_input": (
        "^Replying as ([a-zA-Z0-9_]+). Provide feedback to ([a-zA-Z0-9_]+). Press enter to skip and use auto-reply, or type 'exit' to end the conversation: $",
    ),
    "next_speaker": (
        "^Next speaker: [a-zA-Z0-9_]+$",
        "^\\u001b\\[32m\nNext speaker: [a-zA-Z0-9_]+\n\\u001b\\[0m$",
    ),
}


def _match(key: str, string: str, /) -> bool:
    return any(re.match(pattern, string) for pattern in _patterns[key])


def _findall(key: str, string: str, /) -> tuple[str, ...]:
    for pattern in _patterns[key]:
        if re.match(pattern, string):
            return re.findall(pattern, string)[0]  # type: ignore[no-any-return]
    return ()  # type: ignore[no-any-return]


@dataclass
class CurrentMessage:
    workflow_uuid: str
    sender: Optional[str] = None
    recipient: Optional[str] = None
    type: MessageType = "text_message"
    auto_reply: bool = False
    body: Optional[str] = None
    call_id: Optional[str] = None
    function_name: Optional[str] = None
    arguments: Optional[dict[str, Any]] = None
    retval: Optional[Any] = None

    def process_chunk(self, chunk: str) -> bool:  # noqa: C901
        # logger.info(f"CurrentMessage.process_chunk({chunk=}):")
        if _match("end_of_message", chunk):
            return True

        if _match("auto_reply", chunk):
            # logger.info("CurrentMessage.process_chunk(): auto_reply detected")
            self.auto_reply = True
        elif _match("sender_recipient", chunk):
            # logger.info("CurrentMessage.process_chunk(): sender_recipient detected")
            self.sender, self.recipient = _findall("sender_recipient", chunk)
        elif _match("suggested_function_call", chunk):
            # logger.info("CurrentMessage.process_chunk(): suggested_function_call detected")
            self.call_id, self.function_name = _findall(
                "suggested_function_call", chunk
            )
            self.type = "suggested_function_call"
        elif _match("stars", chunk):
            # logger.info("CurrentMessage.process_chunk(): stars detected")
            pass
        elif _match("function_call_execution", chunk):
            # logger.info("CurrentMessage.process_chunk(): function_call_execution detected")
            self.function_name = _findall("function_call_execution", chunk)  # type: ignore[assignment]
            self.type = "function_call_execution"
        elif _match("response_from_calling_tool", chunk):
            # logger.info("CurrentMessage.process_chunk(): response_from_calling_tool detected")
            self.type = "function_call_execution"
            self.call_id = _findall("response_from_calling_tool", chunk)  # type: ignore[assignment]
        elif _match("no_human_input_received", chunk):
            # logger.info("CurrentMessage.process_chunk(): no_human_input_received detected")
            pass
        elif _match("user_interrupted", chunk):
            # logger.info("CurrentMessage.process_chunk(): user_interrupted detected")
            pass
        elif _match("next_speaker", chunk):
            # logger.info("CurrentMessage.process_chunk(): next_speaker detected")
            pass
        else:
            if self.type == "suggested_function_call":
                if _match("arguments", chunk):
                    # logger.info("CurrentMessage.process_chunk(): parsing arguments")
                    arguments_json: str = _findall("arguments", chunk)  # type: ignore[assignment]
                    self.arguments = json.loads(arguments_json)
                else:
                    logger.warning(
                        f"CurrentMessage.process_chunk(): unexpected chunk: {chunk=}, {self=}"
                    )
            elif self.type == "function_call_execution":
                # logger.info("CurrentMessage.process_chunk(): parsing retval")
                self.retval = chunk
            else:
                # logger.info("CurrentMessage.process_chunk(): parsing body")
                self.body = chunk if self.body is None else self.body + chunk

        return False

    def process_input(
        self, prompt: str, password: bool, messages: list[IOMessage]
    ) -> AskingMessage:
        last_message = messages[-1]
        sender, recipient = None, None
        message: AskingMessage

        if _match("auto_reply_input", prompt):
            # logger.info("IOStreamAdapter.input(): auto_reply_input detected")
            sender, recipient = _findall("auto_reply_input", prompt)  # type: ignore[assignment]

        if last_message.type == "suggested_function_call":
            # logger.info("IOStreamAdapter.input(): suggested_function_call detected")
            message = MultipleChoice(
                sender=sender,
                recipient=recipient,
                prompt="Please approve the suggested function call.",
                choices=["Approve", "Reject", "Exit"],
                default="Approve",
                workflow_uuid=self.workflow_uuid,
            )
        else:
            # logger.info("IOStreamAdapter.input(): text_message detected")
            message = TextInput(
                sender=None,
                recipient=None,
                prompt=prompt,
                password=password,
                workflow_uuid=self.workflow_uuid,
            )

        return message

    def create_message(self) -> list[IOMessage]:
        retval: list[IOMessage] = []
        kwargs = {k: v for k, v in asdict(self).items() if v is not None}
        if (
            kwargs.get("type") in ["suggested_function_call", "function_call_execution"]
            and "body" in kwargs
        ):
            body = kwargs.pop("body")
            sender = kwargs.get("sender")
            recipient = kwargs.get("recipient")
            retval = [
                IOMessage.create(
                    body=body, sender=sender, recipient=recipient, type="text_message"
                )
            ]

        retval.append(IOMessage.create(**kwargs))
        # logger.info(f"CurrentMessage.create_message(): {kwargs=}")
        return retval


class IOStreamAdapter:  # IOStream
    def __init__(self, ui: UI) -> None:
        """Initialize the adapter with a ChatableIO object.

        Args:
            ui (ChatableIO): The ChatableIO object to adapt

        """
        self.ui = ui
        self.current_message = CurrentMessage(ui._workflow_uuid)

        self.messages: list[IOMessage] = []
        # if not isinstance(self.ui, UI):
        #     raise ValueError("The ui object must be an instance of UI.")

    def _process_message_chunk(self, chunk: str) -> int:
        if self.current_message.process_chunk(chunk):
            msgs = self.current_message.create_message()
            for msg in msgs:
                self.messages.append(msg)
            self.current_message = CurrentMessage(self.ui._workflow_uuid)

            return len(msgs)
        else:
            return 0

    def print(
        self, *objects: Any, sep: str = " ", end: str = "\n", flush: bool = False
    ) -> None:
        # logger.info(f"print(): {objects=}, {sep=}, {end=}, {flush=}")
        body = sep.join(map(str, objects)) + end
        num_to_send = self._process_message_chunk(body)
        for i in range(-num_to_send, 0, 1):
            message = self.messages[i]
            self.ui.process_message(message)

    def input(self, prompt: str = "", *, password: bool = False) -> str:
        # logger.info(f"input(): {prompt=}, {password=}")
        message: AskingMessage = self.current_message.process_input(
            prompt, password, self.messages
        )

        retval: str = self.ui.process_message(message)  # type: ignore[assignment]

        # in case of approving a suggested function call, we need to return an empty string to AutoGen
        if (
            message.type == "multiple_choice"
            and self.messages[-1].type == "suggested_function_call"
            and retval == "Approve"
        ):
            retval = ""
        if retval == "Exit":
            retval = "exit"

        # logger.info(f"input(): {retval=}")
        return retval


class AutoGenWorkflows(WorkflowsProtocol):
    def __init__(self) -> None:
        """Initialize the workflows."""
        self._workflows: dict[str, tuple[Callable[[UI, dict[str, Any]], str], str]] = {}

    def register(
        self, name: str, description: str, *, fail_on_redefintion: bool = False
    ) -> Callable[[Workflow], Workflow]:
        def decorator(func: Workflow) -> Workflow:
            check_register_decorator(func)
            if name in self._workflows:
                if fail_on_redefintion:
                    raise ValueError(f"A workflow with name '{name}' already exists.")
                else:
                    logger.warning(f"Overwriting workflow with name '{name}'")

            self._workflows[name] = func, description
            return func

        return decorator

    def run(
        self,
        name: str,
        ui: UI,
        user_id: Optional[str] = None,
        **kwargs: Any,
    ) -> str:
        workflow, _ = self._workflows[name]

        iostream = IOStreamAdapter(ui)

        with IOStream.set_default(iostream):
            # todo: inject user_id into call (and other stuff)
            try:
                ui.workflow_started(
                    sender="AutoGenWorkflows",
                    recipient="User",
                    name=name,
                    description=self.get_description(name),
                    params=kwargs,
                )
                retval = workflow(ui, kwargs)

            except Exception as e:
                logger.error(
                    f"Unhandled exception occurred when executing the workflow: {e}",
                    stack_info=True,
                )
                ui.error(
                    sender="AutoGenWorkflows",
                    recipient="User",
                    short="Unhandled exception occurred when executing the workflow.",
                    long=str(e),
                )
                retval = (
                    f"Unhandled exception occurred when executing the workflow: {e}"
                )

            ui.workflow_completed(
                sender="AutoGenWorkflows",
                recipient="User",
                result=retval,
            )
            logger.info(f"Workflow '{name}' completed with result: {retval}")

            return retval

    @property
    def names(self) -> list[str]:
        return list(self._workflows.keys())

    def get_description(self, name: str) -> str:
        _, description = self._workflows.get(name, (None, "Description not available!"))
        return description

    def register_api(
        self,
        api: "OpenAPI",
        callers: Union[ConversableAgent, Iterable[ConversableAgent]],
        executors: Union[ConversableAgent, Iterable[ConversableAgent]],
        functions: Optional[
            Union[str, Iterable[Union[str, Mapping[str, Mapping[str, str]]]]]
        ] = None,
    ) -> None:
        if not isinstance(callers, Iterable):
            callers = [callers]
        if not isinstance(executors, Iterable):
            executors = [executors]
        if isinstance(functions, str):
            functions = [functions]

        for caller in callers:
            api._register_for_llm(caller, functions=functions)

        for executor in executors:
            api._register_for_execution(executor, functions=functions)


@runtime_checkable
class Toolable(Protocol):
    def register(
        self,
        *,
        caller: ConversableAgent,
        executor: Union[ConversableAgent, list[ConversableAgent]],
    ) -> None: ...



from pathlib import Path
from typing import Any, Callable, Optional

import mesop.labs as mel
import mesop.server.static_file_serving
import mesop.server.wsgi_app
from flask import Flask, Response
from mesop.server.static_file_serving import (
    WEB_COMPONENTS_PATH_SEGMENT,
    noop,
    send_file_compressed,
)
from mesop.server.static_file_serving import (
    configure_static_file_serving as configure_static_file_serving_original,
)

from ...logging import get_logger

logger = get_logger(__name__)


def configure_static_file_serving(
    app: Flask,
    static_file_runfiles_base: str,
    livereload_script_url: Optional[str] = None,
    preprocess_request: Callable[[], None] = noop,
    disable_gzip_cache: bool = False,
    default_allowed_iframe_parents: str = "'self'",
) -> None:
    logger.info("Configuring static file serving with patched method")

    configure_static_file_serving_original(
        app=app,
        static_file_runfiles_base=static_file_runfiles_base,
        livereload_script_url=livereload_script_url,
        preprocess_request=preprocess_request,
        disable_gzip_cache=disable_gzip_cache,
        default_allowed_iframe_parents=default_allowed_iframe_parents,
    )

    @app.route(f"/{WEB_COMPONENTS_PATH_SEGMENT}/__fast_agency_internal__/<path:path>")  # type: ignore[misc]
    def serve_web_components_fast_agency(path: str) -> Response:
        logger.info(f"Serve web components fast agency: {path}")

        root = Path(__file__).parents[3].resolve()
        serving_path = f"{root}/{path}"

        return send_file_compressed(  # type: ignore[no-any-return]
            serving_path,
            disable_gzip_cache=disable_gzip_cache,
        )


logger.info("Patching static file serving in Mesop")
mesop.server.wsgi_app.configure_static_file_serving = configure_static_file_serving


@mel.web_component(path="/__fast_agency_internal__/javascript/wakeup_component.js")  # type: ignore[misc]
def wakeup_component(
    *,
    on_wakeup: Callable[[mel.WebEvent], Any],
    key: Optional[str] = None,
) -> Any:
    return mel.insert_web_component(
        name="wakeup-component",
        key=key,
        events={
            "wakeupEvent": on_wakeup,
        },
        properties={},
    )



import getpass
import json
import textwrap
from collections.abc import Iterator
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, Optional
from uuid import uuid4

from ...base import (
    CreateWorkflowUIMixin,
    Runnable,
)
from ...logging import get_logger
from ...messages import (
    IOMessage,
    MessageProcessorMixin,
    MultipleChoice,
    TextInput,
    TextMessage,
)

logger = get_logger(__name__)


class ConsoleUI(MessageProcessorMixin, CreateWorkflowUIMixin):  # implements UI
    @dataclass
    class ConsoleMessage:
        """A console message."""

        sender: Optional[str]
        recipient: Optional[str]
        heading: Optional[str]
        body: Optional[str]

    def __init__(
        self,
        super_conversation: Optional["ConsoleUI"] = None,
    ) -> None:
        """Initialize the console UI object.

        Args:
            super_conversation (Optional[UI], optional): The super conversation. Defaults to None.
        """
        self.super_conversation: Optional[ConsoleUI] = super_conversation
        self.sub_conversations: list[ConsoleUI] = []

    @contextmanager
    def create(self, app: Runnable, import_string: str) -> Iterator[None]:
        yield

    def start(
        self,
        *,
        app: Runnable,
        import_string: str,
        name: Optional[str] = None,
        params: dict[str, Any],
        single_run: bool = False,
    ) -> None:
        workflow_uuid = uuid4().hex
        ui = self.create_workflow_ui(workflow_uuid=workflow_uuid)
        name = name or app.provider.names[0]
        app.provider.run(name=name, ui=ui, **params)

    @property
    def level(self) -> int:
        return (
            0 if self.super_conversation is None else self.super_conversation.level + 1
        )

    def _format_message(self, console_msg: ConsoleMessage) -> str:
        heading = f"[{console_msg.heading}]" if console_msg.heading else ""
        title = f"{console_msg.sender} -> {console_msg.recipient} {heading}"[:74]

        s = f"""╭─ {title} {'─' * (74 - len(title))}─╮
│
{textwrap.indent(textwrap.fill(console_msg.body if console_msg.body else '', replace_whitespace=False, drop_whitespace=False), '│ ', predicate=lambda line: True)}
╰{'─' * 78}╯
"""
        # remove empty lines
        s = "\n".join([line for line in s.split("\n") if line.strip()])

        # add trailing withespace and │ to each line except the first and the last
        lines = s.split("\n")
        s = (
            lines[0]
            + "\n"
            + "\n".join([line + " " * (79 - len(line)) + "│" for line in lines[1:-1]])
            + "\n"
            + lines[-1]
            + "\n"
        )

        return s

    def _indent(self, text: str) -> str:
        return textwrap.indent(text, " " * 4 * self.level)

    def _format_and_print(self, console_msg: ConsoleMessage) -> None:
        msg = self._format_message(console_msg)
        msg = self._indent(msg)

        print(msg)  # noqa: T201 `print` found

    def visit_default(self, message: IOMessage) -> None:
        content = message.model_dump()["content"]
        console_msg = self.ConsoleMessage(
            sender=message.sender,
            recipient=message.recipient,
            heading=message.type,
            body=json.dumps(content, indent=2),
        )
        self._format_and_print(console_msg)

    def visit_text_message(self, message: TextMessage) -> None:
        console_msg = self.ConsoleMessage(
            sender=message.sender,
            recipient=message.recipient,
            heading=message.type,
            body=message.body,
        )
        self._format_and_print(console_msg)

    def visit_text_input(self, message: TextInput) -> str:
        suggestions = (
            f" (suggestions: {', '.join(message.suggestions)})"
            if message.suggestions
            else ""
        )
        console_msg = self.ConsoleMessage(
            sender=message.sender,
            recipient=message.recipient,
            heading=message.type,
            body=f"{message.prompt}{suggestions}:",
        )

        prompt = self._format_message(console_msg)
        prompt = self._indent(prompt)
        if message.password:
            return getpass.getpass(prompt)
        else:
            return input(prompt)

    def visit_multiple_choice(self, message: MultipleChoice) -> str:
        console_msg = self.ConsoleMessage(
            sender=message.sender,
            recipient=message.recipient,
            heading=message.type,
            body=f"{message.prompt} (choices: {', '.join(message.choices)}, default: {message.default})",
        )

        prompt = self._format_message(console_msg)
        prompt = self._indent(prompt)
        while True:
            # logger.info(f"visit_multiple_choice(): {prompt=}")
            retval = input(prompt)
            if retval in message.choices:
                return retval
            elif retval == "" and message.default:
                return message.default
            else:
                print(f"Invalid choice ('{retval}'). Please try again.")  # noqa: T201 `print` found

    def process_message(self, message: IOMessage) -> Optional[str]:
        # logger.info(f"process_message(): {message=}")
        return self.visit(message)

    # def process_streaming_message(self, message: IOStreamingMessage) -> str | None:
    #     raise NotImplementedError

    def create_subconversation(self) -> "ConsoleUI":
        sub_conversation = ConsoleUI(self)
        self.sub_conversations.append(sub_conversation)

        return sub_conversation



import time
from collections.abc import Iterator
from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable, Optional

import mesop as me

from fastagency.base import ProviderProtocol

from ...logging import get_logger
from .data_model import Conversation, State
from .message import consume_responses, message_box
from .send_prompt import send_prompt_to_autogen
from .styles import MesopHomePageStyles

if TYPE_CHECKING:
    from .mesop import MesopUI

__all__ = ["me"]

# Get the logger
logger = get_logger(__name__)


DEFAULT_SECURITY_POLICY = me.SecurityPolicy(
    allowed_script_srcs=["https://cdn.jsdelivr.net"],
    allowed_iframe_parents=["https://fastagency.ai"],
)


def create_home_page(
    ui: "MesopUI",
    *,
    styles: Optional[MesopHomePageStyles] = None,
    security_policy: Optional[me.SecurityPolicy] = None,
) -> Callable[[], None]:
    mhp = MesopHomePage(ui, styles=styles, security_policy=security_policy)

    return mhp.build()


@dataclass
class MesopHomePageParams:
    # header_title: str = "FastAgency - Mesop"
    conv_starter_text: str = "Select workflow to use with FastAgency team"
    no_workflows_text: str = "No workflows found, click to retry"
    workflows_exception_text: str = (
        "An exception occurred while discovering workflows, click to retry"
    )


class MesopHomePage:
    def __init__(
        self,
        ui: "MesopUI",
        *,
        params: Optional[MesopHomePageParams] = None,
        styles: Optional[MesopHomePageStyles] = None,
        security_policy: Optional[me.SecurityPolicy] = None,
    ) -> None:
        self._ui = ui
        self._params = params or MesopHomePageParams()
        self._styles = styles or MesopHomePageStyles()
        self._security_policy = security_policy or DEFAULT_SECURITY_POLICY

    def build(self) -> Callable[[], None]:
        @me.page(  # type: ignore[misc]
            path="/",
            stylesheets=self._styles.stylesheets,
            security_policy=self._security_policy,
        )
        def home_page() -> None:
            self.home_page()

        return home_page  # type: ignore[no-any-return]

    def home_page(self) -> None:
        try:
            state = me.state(State)
            with me.box(style=self._styles.root):
                self.past_conversations_box()
                if state.in_conversation:
                    self.conversation_box()
                else:
                    self.conversation_starter_box()
        except Exception as e:
            logger.error(f"home_page(): Error rendering home page: {e}")
            me.text(text="Error: Something went wrong, please check logs for details.")

    def header(self) -> None:
        with me.box(
            style=self._styles.header,
        ):
            me.text(
                self._ui.app.title,
                style=self._styles.header_text,
            )

    def past_conversations_box(self) -> None:
        def conversation_display_title(full_name: str, max_length: int) -> str:
            if len(full_name) <= max_length:
                return full_name
            else:
                return full_name[: max_length - 3] + "..."

        def select_past_conversation(ev: me.ClickEvent) -> Iterator[None]:
            id = ev.key
            state = me.state(State)
            conversations_with_id = list(
                filter(lambda c: c.id == id, state.past_conversations)
            )
            conversation = conversations_with_id[0]
            state.conversation = conversation
            state.in_conversation = True
            yield
            time.sleep(1)
            yield
            me.scroll_into_view(key="end_of_messages")
            yield

        def on_show_hide(ev: me.ClickEvent) -> None:
            state.hide_past = not state.hide_past

        def on_start_new_conversation(ev: me.ClickEvent) -> None:
            state.in_conversation = False
            state.prompt = ""

        state = me.state(State)
        style = (
            self._styles.past_chats_hide
            if state.hide_past
            else self._styles.past_chats_show
        )
        with me.box(style=style):
            with me.box(
                style=self._styles.past_chats_inner,
            ):
                with me.content_button(
                    on_click=on_show_hide, disabled=not state.past_conversations
                ):
                    me.icon("menu")
                with me.content_button(
                    on_click=on_start_new_conversation,
                    disabled=not state.conversation.completed,
                ):
                    me.icon("rate_review")
            if not state.hide_past:
                for conversation in state.past_conversations:
                    with me.box(
                        key=conversation.id,  # they are GUIDs so should not clash with anything other on the page
                        on_click=select_past_conversation,
                        style=self._styles.past_chats_conv,
                    ):
                        me.text(
                            text=conversation_display_title(conversation.title, 128)
                        )

    def conversation_starter_box(self) -> None:
        def retry(ev: me.ClickEvent) -> None:
            state = me.state(State)
            try:
                state.available_workflows = provider.names
                state.available_workflows_exception = False
            except Exception:
                state.available_workflows_exception = False

        provider = self.get_provider()
        with me.box(style=self._styles.chat_starter):
            self.header()
            with me.box(
                style=self._styles.conv_starter,
            ):
                me.text(
                    self._params.conv_starter_text,
                    style=self._styles.conv_starter_text,
                )
                with me.box(style=self._styles.conv_starter_wf_box):
                    state = me.state(State)
                    if not state.available_workflows_initialized:
                        state.available_workflows_initialized = True
                        try:
                            state.available_workflows = provider.names
                            state.available_workflows_exception = False
                        except Exception:
                            state.available_workflows = []
                            state.available_workflows_exception = True

                    names = state.available_workflows
                    if names and not state.available_workflows_exception:
                        try:
                            for wf_name in names:
                                wf_description = provider.get_description(wf_name)
                                with me.content_button(
                                    key=wf_name, on_click=lambda e: self.send_prompt(e)
                                ):
                                    me.text(wf_description)
                        except Exception:
                            state.available_workflows_exception = True
                    if not names or state.available_workflows_exception:
                        with me.content_button(on_click=retry):
                            if state.available_workflows_exception:
                                me.text(self._params.workflows_exception_text)
                            else:
                                me.text(self._params.no_workflows_text)

    def get_provider(self) -> ProviderProtocol:
        ui = self._ui
        return ui.app.provider

    def send_prompt(self, ev: me.ClickEvent) -> Iterator[None]:
        name = ev.key
        provider = self.get_provider()
        state = me.state(State)
        conversation = Conversation(
            title="New Conversation", completed=False, waiting_for_feedback=False
        )
        state.conversation = conversation
        state.in_conversation = True
        yield
        responses = send_prompt_to_autogen(provider=provider, name=name)
        yield from consume_responses(responses)
        try:
            state.available_workflows = provider.names
        except Exception:
            state.available_workflows = []
            state.available_workflows_exception = True

    def conversation_box(self) -> None:
        state = me.state(State)
        conversation = state.conversation
        with me.box(style=self._styles.chat_starter):
            self.header()
            messages = conversation.messages
            with me.box(
                style=self._styles.conv_list,
            ):
                me.box(
                    key="conversationtop",
                    style=self._styles.conv_top,
                )
                for message in messages:
                    message_box(
                        message, conversation.is_from_the_past, styles=self._styles
                    )
                if messages:
                    me.box(
                        key="end_of_messages",
                        style=self._styles.conv_top,
                    )



from ...helpers import check_imports

check_imports(["fastapi_code_generator", "fastapi", "requests"], "openapi")

from .client import OpenAPI  # noqa: E402
from .patch_fastapi_code_generator import patch_parse_schema  # noqa: E402

patch_parse_schema()

__all__ = ["OpenAPI"]



from .base import (
    JETSTREAM,
    InitiateWorkflowModel,
    InputResponseModel,
    NatsAdapter,
    NatsProvider,
)

__all__ = [
    "NatsAdapter",
    "InitiateWorkflowModel",
    "InputResponseModel",
    "JETSTREAM",
    "NatsProvider",
]



def darken_hex_color(hex_color: str, factor: float = 0.8) -> str:
    """Darkens a hex color by a given factor.

    Args:
    hex_color: The hex color code (e.g., '#FF0000').
    factor: The darkening factor (0.0 to 1.0, where 1.0 is no change and 0.0 is completely dark).

    Returns:
    The darkened hex color code.
    """
    # Remove the '#' prefix if it exists
    hex_color = hex_color.lstrip("#")

    if len(hex_color) == 3:
        hex_color = "".join(char * 2 for char in hex_color)

    # Convert hex to RGB values
    rgb = tuple(int(hex_color[i : i + 2], 16) for i in (0, 2, 4))

    # Darken each component
    darkened_rgb = tuple(int(channel * factor) for channel in rgb)

    # Convert back to hex
    darkened_hex = "#{:02X}{:02X}{:02X}".format(*darkened_rgb)

    return darkened_hex



__all__ = ["FastAgency"]

from collections.abc import Awaitable, Generator
from contextlib import contextmanager
from typing import Any, Callable, Optional, Union

from .base import (
    ASGIProtocol,
    ProviderProtocol,
    Runnable,
    UIBase,
    WSGIProtocol,
)
from .exceptions import (
    FastAgencyASGINotImplementedError,
    FastAgencyWSGINotImplementedError,
)
from .logging import get_logger

logger = get_logger(__name__)


class FastAgency:  # Runnable
    def __init__(
        self,
        provider: ProviderProtocol,
        ui: UIBase,
        *,
        title: Optional[str] = None,
        description: Optional[str] = None,
    ) -> None:
        """Initialize the FastAgency object.

        Args:
            provider (ProviderProtocol): The provider object to use
            ui (UI): The UI object to use
            title (Optional[str], optional): The title of the FastAgency. If None, the default string will be used. Defaults to None.
            description (Optional[str], optional): The description of the FastAgency. If None, the default string will be used. Defaults to None.
        """
        _self: Runnable = self
        self._title = title or "FastAgency application"
        self._description = description or "FastAgency application"

        logger.info(
            f"Initializing FastAgency {self} with workflows: {provider} and UI: {ui}"
        )
        self._provider = provider
        self._ui = ui
        logger.info(f"Initialized FastAgency: {self}")

    @property
    def title(self) -> str:
        """Return the title of the FastAgency."""
        return self._title

    @property
    def description(self) -> str:
        """Return the description of the FastAgency."""
        return self._description

    def __str__(self) -> str:
        """Return the string representation of the FastAgency."""
        return f"<FastAgency title={self._title}>"

    @property
    def provider(self) -> ProviderProtocol:
        """Return the provider object."""
        return self._provider

    @property
    def ui(self) -> UIBase:
        """Return the UI object."""
        return self._ui

    @contextmanager
    def create(self, import_string: str) -> Generator[None, None, None]:
        """Create the FastAgency."""
        with self._ui.create(app=self, import_string=import_string):
            yield

    def start(
        self,
        *,
        import_string: str,
        name: Optional[str] = None,
        params: dict[str, Any],
        single_run: bool = False,
    ) -> None:
        """Start the FastAgency."""
        self.ui.start(
            app=self,
            import_string=import_string,
            name=name,
            params=params,
            single_run=single_run,
        )

    def __call__(self, *args: Any) -> Union[Awaitable[None], list[bytes]]:
        if len(args) == 2 and callable(args[1]):
            # WSGI interface
            environ, start_response = args
            return self.handle_wsgi(environ, start_response)
        elif len(args) == 3 and callable(args[1]) and callable(args[2]):
            # ASGI interface
            scope, receive, send = args
            scope_type = scope.get("type")
            if scope_type == "http":
                return self.handle_asgi(scope, receive, send)
            else:
                raise NotImplementedError(
                    f"ASGI scope type '{scope_type}' not supported."
                )
        else:
            raise TypeError(f"Invalid arguments for __call__: {args}")

    def handle_wsgi(
        self, environ: dict[str, Any], start_response: Callable[..., Any]
    ) -> list[bytes]:
        logger.debug(f"Handling WSGI request: {environ}")
        if isinstance(self.ui, WSGIProtocol):
            return self.ui.handle_wsgi(self, environ, start_response)
        else:
            raise FastAgencyWSGINotImplementedError(
                "WSGI interface not supported for UI: {self.ui}"
            )

    async def handle_asgi(
        self,
        scope: dict[str, Any],
        receive: Callable[[dict[str, Any]], Awaitable[None]],
        send: Callable[[dict[str, Any]], Awaitable[None]],
    ) -> None:
        if isinstance(self.ui, ASGIProtocol):
            return await self.ui.handle_asgi(self, scope, receive, send)
        else:
            raise FastAgencyASGINotImplementedError(
                "ASGI interface not supported for UI: {self.ui}"
            )



"""The fastest way to bring multi-agent workflows to production."""

from .__about__ import __version__
from .app import FastAgency
from .base import UI, UIBase, Workflow, WorkflowsProtocol
from .messages import (
    FunctionCallExecution,
    IOMessage,
    MessageType,
    MultipleChoice,
    SuggestedFunctionCall,
    TextInput,
    TextMessage,
)

__all__ = [
    "UIBase",
    "FastAgency",
    "FunctionCallExecution",
    "IOMessage",
    "MessageType",
    "MultipleChoice",
    "SuggestedFunctionCall",
    "TextInput",
    "TextMessage",
    "WorkflowsProtocol",
    "Workflow",
    "UI",
    "__version__",
]



import logging
from typing import Any, ClassVar, Literal, Optional, Protocol

import requests
from pydantic import BaseModel, model_validator
from typing_extensions import TypeAlias

# Get the logger
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

BaseSecurityType: TypeAlias = type["BaseSecurity"]


class BaseSecurity(BaseModel):
    """Base class for security classes."""

    type: ClassVar[Literal["apiKey", "http", "mutualTLS", "oauth2", "openIdConnect"]]
    in_value: ClassVar[Literal["header", "query", "cookie", "bearer", "basic", "tls"]]
    name: str

    @model_validator(mode="after")  # type: ignore[misc]
    def __post_init__(
        self,
    ) -> None:  # dataclasses uses __post_init__ instead of model_validator
        """Validate the in_value based on the type."""
        valid_in_values = {
            "apiKey": ["header", "query", "cookie"],
            "http": ["bearer", "basic"],
            "oauth2": ["bearer"],
            "openIdConnect": ["bearer"],
            "mutualTLS": ["tls"],
        }
        if self.in_value not in valid_in_values[self.type]:
            raise ValueError(
                f"Invalid in_value '{self.in_value}' for type '{self.type}'"
            )

    def accept(self, security_params: "BaseSecurityParameters") -> bool:
        return isinstance(self, security_params.get_security_class())

    @classmethod
    def is_supported(cls, type: str, schema_parameters: dict[str, Any]) -> bool:
        return cls.type == type and cls.in_value == schema_parameters.get("in")

    @classmethod
    def get_security_class(
        cls, type: str, schema_parameters: dict[str, Any]
    ) -> BaseSecurityType:
        sub_classes = cls.__subclasses__()

        for sub_class in sub_classes:
            if sub_class.is_supported(type, schema_parameters):
                return sub_class
        else:
            logger.error(
                f"Unsupported type '{type}' and schema_parameters '{schema_parameters}' combination"
            )
            raise ValueError(
                f"Unsupported type '{type}' and schema_parameters '{schema_parameters}' combination"
            )

    @classmethod
    def get_security_parameters(cls, schema_parameters: dict[str, Any]) -> str:
        return f"{cls.__name__}(name=\"{schema_parameters.get('name')}\")"


class BaseSecurityParameters(Protocol):
    """Base class for security parameters."""

    def apply(
        self,
        q_params: dict[str, Any],
        body_dict: dict[str, Any],
        security: BaseSecurity,
    ) -> None: ...

    def get_security_class(self) -> type[BaseSecurity]: ...


class APIKeyHeader(BaseSecurity):
    """API Key Header security class."""

    type: ClassVar[Literal["apiKey"]] = "apiKey"
    in_value: ClassVar[Literal["header"]] = "header"

    class Parameters(BaseModel):  # BaseSecurityParameters
        """API Key Header security parameters class."""

        value: str

        def apply(
            self,
            q_params: dict[str, Any],
            body_dict: dict[str, Any],
            security: BaseSecurity,
        ) -> None:
            api_key_header: APIKeyHeader = security  # type: ignore[assignment]

            if "headers" not in body_dict:
                body_dict["headers"] = {}

            body_dict["headers"][api_key_header.name] = self.value

        def get_security_class(self) -> type[BaseSecurity]:
            return APIKeyHeader


class APIKeyQuery(BaseSecurity):
    """API Key Query security class."""

    type: ClassVar[Literal["apiKey"]] = "apiKey"
    in_value: ClassVar[Literal["query"]] = "query"

    @classmethod
    def is_supported(cls, type: str, schema_parameters: dict[str, Any]) -> bool:
        return (
            super().is_supported(type, schema_parameters)
            and "name" in schema_parameters
        )

    class Parameters(BaseModel):  # BaseSecurityParameters
        """API Key Query security parameters class."""

        value: str

        def apply(
            self,
            q_params: dict[str, Any],
            body_dict: dict[str, Any],
            security: BaseSecurity,
        ) -> None:
            api_key_query: APIKeyQuery = security  # type: ignore[assignment]

            q_params[api_key_query.name] = self.value

        def get_security_class(self) -> type[BaseSecurity]:
            return APIKeyQuery


class APIKeyCookie(BaseSecurity):
    """API Key Cookie security class."""

    type: ClassVar[Literal["apiKey"]] = "apiKey"
    in_value: ClassVar[Literal["cookie"]] = "cookie"

    class Parameters(BaseModel):  # BaseSecurityParameters
        """API Key Cookie security parameters class."""

        value: str

        def apply(
            self,
            q_params: dict[str, Any],
            body_dict: dict[str, Any],
            security: BaseSecurity,
        ) -> None:
            api_key_cookie: APIKeyCookie = security  # type: ignore[assignment]

            if "cookies" not in body_dict:
                body_dict["cookies"] = {}

            body_dict["cookies"][api_key_cookie.name] = self.value

        def get_security_class(self) -> type[BaseSecurity]:
            return APIKeyCookie


class HTTPBearer(BaseSecurity):
    """HTTP Bearer security class."""

    type: ClassVar[Literal["http"]] = "http"
    in_value: ClassVar[Literal["bearer"]] = "bearer"

    class Parameters(BaseModel):  # BaseSecurityParameters
        """HTTP Bearer security parameters class."""

        value: str

        def apply(
            self,
            q_params: dict[str, Any],
            body_dict: dict[str, Any],
            security: BaseSecurity,
        ) -> None:
            if "headers" not in body_dict:
                body_dict["headers"] = {}

            body_dict["headers"]["Authorization"] = f"Bearer {self.value}"

        def get_security_class(self) -> type[BaseSecurity]:
            return HTTPBearer


class OAuth2PasswordBearer(BaseSecurity):
    """OAuth2 Password Bearer security class."""

    type: ClassVar[Literal["oauth2"]] = "oauth2"
    in_value: ClassVar[Literal["bearer"]] = "bearer"
    token_url: str

    @classmethod
    def is_supported(cls, type: str, schema_parameters: dict[str, Any]) -> bool:
        return type == cls.type and "password" in schema_parameters.get("flows", {})

    @classmethod
    def get_security_parameters(cls, schema_parameters: dict[str, Any]) -> str:
        name = schema_parameters.get("name")
        token_url = f'{schema_parameters.get("server_url")}/{schema_parameters["flows"]["password"]["tokenUrl"]}'
        return f'{cls.__name__}(name="{name}", token_url="{token_url}")'

    class Parameters(BaseModel):  # BaseSecurityParameters
        """OAuth2 Password Bearer security class."""

        username: Optional[str] = None
        password: Optional[str] = None
        bearer_token: Optional[str] = None
        token_url: Optional[str] = None

        @model_validator(mode="before")
        def check_credentials(cls, values: dict[str, Any]) -> Any:  # noqa
            username = values.get("username")
            password = values.get("password")
            bearer_token = values.get("bearer_token")

            if not bearer_token and (not username or not password):
                # If bearer_token is not provided, both username and password must be defined
                raise ValueError(
                    "Both username and password are required if bearer_token is not provided."
                )

            return values

        def get_token(self, token_url: str) -> str:
            # Get the token
            request = requests.post(
                token_url,
                data={
                    "username": self.username,
                    "password": self.password,
                },
                timeout=5,
            )
            request.raise_for_status()
            return request.json()["access_token"]  # type: ignore

        def apply(
            self,
            q_params: dict[str, Any],
            body_dict: dict[str, Any],
            security: BaseSecurity,
        ) -> None:
            if not self.bearer_token:
                if security.token_url is None:  # type: ignore
                    raise ValueError("Token URL is not defined")
                self.bearer_token = self.get_token(security.token_url)  # type: ignore

            if "headers" not in body_dict:
                body_dict["headers"] = {}

            body_dict["headers"]["Authorization"] = f"Bearer {self.bearer_token}"

        def get_security_class(self) -> type[BaseSecurity]:
            return OAuth2PasswordBearer



class FastAgencyError(Exception):
    pass


class FastAgencyCLIError(FastAgencyError):
    pass


class FastAgencyCLIPythonVersionError(FastAgencyCLIError):
    pass


class FastAgencyWSGINotImplementedError(FastAgencyError):
    pass


class FastAgencyASGINotImplementedError(FastAgencyError):
    pass


class FastAgencyConnectionError(FastAgencyError):
    pass


class FastAgencyNATSConnectionError(FastAgencyConnectionError):
    pass


class FastAgencyFastAPIConnectionError(FastAgencyConnectionError):
    pass


class FastAgencyKeyError(KeyError):
    pass


class FastAgencyNATSKeyError(FastAgencyKeyError):
    pass


class FastAgencyFastAPIKeyError(FastAgencyKeyError):
    pass



from dataclasses import dataclass, field
from typing import Optional

import mesop as me


@dataclass
class ConversationMessage:
    io_message_json: str = ""
    level: int = 0
    conversation_id: str = ""
    feedback: list[str] = field(default_factory=list)
    feedback_completed: bool = False


@dataclass
class Conversation:
    id: str = ""
    title: str = ""
    completed: bool = False
    waiting_for_feedback: bool = False
    feedback: str = ""
    is_from_the_past: bool = False
    messages: list[ConversationMessage] = field(default_factory=list)
    fastagency: Optional[str] = None


@me.stateclass
class State:
    in_conversation: bool = False  # True when in active conversation, or past one.
    prompt_input: str = ""
    prompt_output: str = ""
    conversation: Conversation
    past_conversations: list[Conversation] = field(default_factory=list)
    hide_past: bool = True
    available_workflows: list[str] = field(default_factory=list)
    available_workflows_initialized = False
    available_workflows_exception = False



import importlib
import json
import tempfile
from collections.abc import Iterator
from contextlib import contextmanager
from json import JSONDecoder
from pathlib import Path
from typing import Optional

__all__ = ["check_imports"]


def check_imports(package_names: list[str], target_name: str) -> None:
    not_importable = [
        f"'{package_name}'"
        for package_name in package_names
        if importlib.util.find_spec(package_name) is None
    ]
    if len(not_importable) > 0:
        raise ImportError(
            f"Package(s) {', '.join(not_importable)} not found. Please install it with:\n\npip install \"fastagency[{target_name}]\"\n"
        )


# based on https://stackoverflow.com/questions/61380028/how-to-detect-and-indent-json-substrings-inside-longer-non-json-text/61384796#61384796
def extract_json_objects(
    text: str, decoder: Optional[JSONDecoder] = None
) -> Iterator[str]:
    decoder = decoder or JSONDecoder()
    pos = 0
    while True:
        # print(f"matching: {text[pos:]}")
        match = text.find("{", pos)
        if match == -1:
            yield text[pos:]  # return the remaining text
            break
        yield text[pos:match].rstrip(" ")  # modification for the non-JSON parts
        try:
            result, index = decoder.raw_decode(text[match:])
            yield result
            pos = match + index
            # move past space characters if needed
            while pos < len(text) and text[pos] == " ":
                pos += 1
        except ValueError:
            yield text[match]
            pos = match + 1


def jsonify_string(line: str) -> str:
    line_parts: list[str] = []
    for result in extract_json_objects(line):
        if isinstance(result, dict):  # got a JSON obj
            line_parts.append(f"\n```\n{json.dumps(result, indent=4)}\n```\n")
        else:  # got text/non-JSON-obj
            line_parts.append(result)
    # (don't make that a list comprehension, quite un-readable)

    return "".join(line_parts)


@contextmanager
def optional_temp_path(path: Optional[str] = None) -> Iterator[Path]:
    if path is None:
        with tempfile.TemporaryDirectory() as temp_dir:
            yield Path(temp_dir)
    else:
        yield Path(path)
